/**********************************************************************
 *
 * sqliteclass.h -- 
 * Copyright (C) 2008  DL Consulting Ltd
 *
 * A component of the Greenstone digital library software
 * from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *********************************************************************/

#ifndef SQLITECLASS_H
#define SQLITECLASS_H


#include "dbclass.h"
#include "sqlite3.h"


class sqliteclass : public dbclass
{
public:
  sqliteclass() { sqlitefile = NULL; }
  ~sqliteclass();

  // returns true if opened
  bool opendatabase (const text_t &filename, int mode, int num_retrys, 
		     bool need_filelock);
  void closedatabase ();

  // returns true on success
  bool getinfo (const text_t& key, infodbclass &info);

  // returns true if exists
  bool exists (const text_t& key);

  // returns true on success
  bool setinfo (const text_t &key, const infodbclass &info);

  void deletekey (const text_t &key);

  // getfirstkey and getnextkey are used for traversing the database
  // no insertions or deletions should be carried out while traversing
  // the database. when there are no keys left to visit in the database
  // an empty string is returned.
  text_t getfirstkey ();
  text_t getnextkey (const text_t &key);

protected:
  text_t openfile;
  sqlite3* sqlitefile;

  bool getinfoline (text_t::iterator &here, text_t::iterator end, 
		    text_t &key, text_t &value);

  bool sqlgetarray (const text_t &sql_cmd, vector<text_tmap> &sql_results);
  bool sqlexec (const text_t &sql_cmd);
};


#endif
